/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.runtime.java;


import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.ImagePointer;




/**
 * <p>Represents the underlying monitor used by a Java Virtual Machine to manage locking and synchronization of a Java object.</p>
 * 
 * <p>The underlying monitor is implementation specific. Some implementations may choose to use their monitor implementations to
 * control access to Java Virtual Machine resources that are not objects. In such cases, {@link #getObject()} will return null.
 * </p>
 * 
 * <p>Java programmers use the synchronized modifier on methods and the synchronized block within methods to control simultaneous access to Java objects.
 * Java uses monitors for this synchronization, which can be implemented using a variety of techniques. The JavaMonitor class presents 
 * the simple monitor abstraction that allows the caller to determine:
 * <ul>
 * <li>Which thread currently owns the monitor</li>
 * <li>Which threads are waiting to be woken after they have gotten ownship of the monitor and relinquished it, normally within 
 *     Object.wait() within a synchronized block or method.</li>
 * <li>The threads that waiting to get ownership of the monitor. These are typically threads waiting to enter a 
 * synchronized block or method.</li>    
 * </ul>
 * </p>
 * 
 * <p>This API presents only what exists at the Java Virtual Machine bytecode level. The locking facilities provided by the {@link java.util.concurrent.lock}
 * package are expected to be implemented on top of ordinary Java monitors.</p>
 */
public interface JavaMonitor {
    
    /**
     * <p>Get the object associated with this monitor. Not all JavaMonitors will have objects, as there may be JavaMonitors that
     * are used to control access to internal Java Virtual Machine resources ("Raw" monitors).
     * </p>
     *  
     * @return the Java object associated with this monitor, or null.
     */
    JavaObject getObject();

    /**
     * <p>Get the name of a monitor.</p>
     * 
     * <p>For monitors not associated with object ("raw" monitors), it is expected that this method will return
     * a descriptive name that is meaningful to the Java Virtual Machine implementation. For example "Heap lock" might
     * be a monitor controlling exclusive access to the Java heap.</p> 
     * 
     * <p>For objects, the expectation is that the name will uniquely identify the object the monitor is associated with.
     * This is not expected to necessarily be consistent between different dumps of the same JVM.
     * </p>
     * 
     * @return the name of the monitor (never null)
     * @throws CorruptDataException 
     */
    String getName() throws CorruptDataException;
    
    /**
     * <p>Get the thread which currently owns the monitor. This may be null if the monitor is not owned.
     * </p>
     * 
     * @return the owner of the monitor, or null if the monitor is not owned
     * @throws CorruptDataException 
     */
    JavaThread getOwner() throws CorruptDataException;
    
    /**
     * <p>Get the set of threads waiting to enter the monitor. 
     * </p>
     * <p>The returned list follows the standard semantics for javax.tools.diagnostics collections.</p>
     * <p>The returned value is never null but can be an empty list.</p>
    
     * @return a list of threads waiting to enter this monitor
     * 
     * @see JavaThread
     * @see javax.tools.diagnostics.image.CorruptData
     */
    List<JavaThread> getEnterWaiters();
    
    /**
     * <p>Get the set of threads waiting to be notified on the monitor. 
     * They are usually threads in the {@link java.lang.Object#wait()} method.</p>
     * <p>The returned list follows the standard semantics for javax.tools.diagnostics collections.</p>
     * <p>The returned value is never null but can be an empty list.</p>
     * 
     * @return a list of threads waiting to be notified on this monitor.
     * 
     * @see JavaThread
     * @see javax.tools.diagnostics.image.CorruptData
     */
    List<JavaThread> getNotifyWaiters();
    
    /**
     * <p>Get the identifier for this monitor.</p>
     * 
     * @return The pointer which uniquely identifies this monitor in memory.
     */
    ImagePointer getID();
    
	/**
	 * @param obj
	 * @return true if the given object refers to the same Java Monitor in the image
	 */
	public boolean equals(Object obj);
	public int hashCode();
}
