/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.runtime.java;


import java.io.File;
import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.image.MemoryAccessException;
import javax.tools.diagnostics.runtime.ManagedRuntime;



/**
 * Represents an instance of a Java Virtual Machine
 * 
 * This interface defines attributes and features common across real implementation of the Java Virtual Machine.
 * Not all of these characteristics are defined by the <a href="http://java.sun.com/docs/books/jvms/second_edition/html/VMSpecTOC.doc.html">Java Virtual Machine Specification</a>
 * 
 * Notable additions beyond the JVM specification include Garbage Collection and access to the contents of the Heap or Heaps.
 *  
 * Since this interface defines a view of the Java Runtime that is beyond that seen by the Java programmer during program execution it is necessarily more detailed.
 *    
 */
public interface JavaRuntime extends ManagedRuntime {

    /**
     * Get the object that represents the virtual machine
     * @return the address of the JavaVM structure which represents this JVM instance in JNI
     * 
     * @throws CorruptDataException 
     */
    ImagePointer getJavaVM() throws CorruptDataException;
    
    /**
     * Fetch the {@link JavaVMInitArgs} which were used to create this VM.
     * See <a href="http://java.sun.com/javase/6/docs/technotes/guides/jni/spec/invocation.html#wp16334">JNI_CreateJavaVM</a> in the JNI Specification for more details.
     * 
     * A valid object is returned or an exception is thrown.
     * 
     * @return the JavaVMInitArgs which were used to create this VM.
     * @throws DataUnavailable if the arguments are not available 
     * @throws CorruptDataException if the implementation was unexpectedly unable to retrieve the data
     */
    JavaVMInitArgs getJavaVMInitArgs() throws DataUnavailable, CorruptDataException;
    
    /**
     * Get the set of class loaders available in this Java Virtual Machine instance.
     * <p>
     * Available in this context means class loaders that are participating in the class loader hierarchy.
     * All class loaders are returned including any defined by the Java Virtual machine instance itself.
     * </p><p>
     * Any structural relationships between class loaders in this list is not exposed. Recreation of the class loader
     * graph within a Java Virtual Machine instance is beyond the scope of the API.  
     * </p>
     * <p>At least one class loader must be returned in the resulting list.</p> 
     * 
     * The returned list follows the standard semantics for javax.tools.diagnostics collections
     * 
     * @return a {@link java.util.List} of all of the class loaders within this Java Virtual Machine instance
     * 
     * @see JavaClassLoader
     */
    List<JavaClassLoader> getJavaClassLoaders();
    
    /**
     * Get the set of Java Threads that have been started {@link java.lang.Thread#start()} in this Java Virtual Machine instance.
     * This method does not return all instances of java.lang.Thread contained within the system.  Only threads that have been started and have 
     * not yet stopped or exited are returned. Threads may not be in an active state when returned by this method.    
     * 
     * The returned list follows the standard semantics for javax.tools.diagnostics collections   
     * 
     * The returned list is never null although it can be empty.
     * 
     * @return a {@link java.util.List} of the JavaThreads in the runtime
     * 
     * @see JavaThread
     */
    
    List<JavaThread> getThreads();
    
    /**
     * Get the set of JavaMethod objects that have been compiled.  
     * 
     * Compiled methods are methods that have been converted into native code by the Java Virtual Machine or related Just In Time Compiler
     * 
     * There is no expectation that any method has been compiled. The returned list could be empty.
     * However any JavaMethod reachable though the API which would return a non empty list for calls to {@link JavaMethod#getCompiledSections()} must be contained within the list
     * returned by this method.  
     * 
     * The returned list follows the standard semantics for javax.tools.diagnostics collections
     * 
     * @return a {@link java.util.List} of all of the JavaMethods in the JavaRuntime which
     * have been compiled.
     * 
     * @see JavaMethod
     */
    List<JavaMethod> getCompiledMethods();
    
    /**
     * Provides access to the collection of monitors used in the Java Virtual Machine. This 
     * collection can include  monitors associated with managed objects (e.g. object
     * monitors) and monitors associated with Java Virtual Machine internal control structures (e.g.
     * raw monitors).   Raw monitors are implementation specific.   
     * 
     *   
     * The returned list follows the standard semantics for javax.tools.diagnostics collections.
     * 
     * The returned list is never null but could be empty.
     * 
     * @return a list of monitors
     * 
     * @see JavaMonitor
     * @see javax.tools.diagnostics.image.CorruptData
     */
    List<JavaMonitor> getMonitors();
    
    /**
     * Get the set of heaps known by the Java Virtual Machine
     * There may be multiple heaps	within a Java Virtual Machine, for instance a generational heap and a class heap. 
     * Heaps may be specific to this Java Virtual Machine instance, or may be shared between multiple
     * Java Virtual Machine instances. 
     *<p>
     * The returned list follows the standard semantics for javax.tools.diagnostics collections.
     * <p>
     * The returned list is never null and will always contain at least one JavaHeap object
     * 
  	 * @return a list for all of the Java heaps within this runtime. 
     * @see JavaHeap
     * @see javax.tools.diagnostics.image.CorruptData
     */
    List<JavaHeap> getHeaps();
    
    /**
     * Get the complete set of object and class roots known to the Java Virtual Machine
     * 
     * @return a list of JavaReferences representing the known global heap roots within this runtime. 
     *
     * The returned list follows the standard semantics for javax.tools.diagnostics collections.
     * 
     * The returned list is never null but can be empty.

     * @see JavaReference
     * @see JavaStackFrame
     * @see javax.tools.diagnostics.image.CorruptData
     * 
     */
    List<JavaReference> getHeapRoots();
    
    /**
     * @remove candidate for removal as its an arbitary implementation specific method
     * 
     * Fetches implementation specific trace buffers, like the verbose GC buffer.
     * 
     * @param bufferName a String naming the buffer to be fetched
     * @param formatted true if formatting should be performed on the buffer, or 
     * false if the raw buffer contents should be returned
     * @return an implementation specific result, depending on the parameters 
     * @throws CorruptDataException 
     */
    Object getTraceBuffer(String bufferName, boolean formatted) throws CorruptDataException;
    
    /**
     * Gets the object located at address <code>address</code> in the heap.
     * @param	addr the <code>ImagePointer</code> instance representing the start address of object in the heap;
     * @return	the <code>JavaObject</code> instance representing the located object.
     * @throws 	IllegalArgumentException	if  <code>address</code> is null, outside the heap's boundaries  or if it doesn't point to the start location of an object;
     * @throws 	MemoryAccessException 		if  <code>address</code> is is in the heap but it's not accessible from the dump;
     * @throws 	CorruptDataException 		if any data needed to build the returned instance of <code>JavaObject</code> is corrupt.
     * @throws 	DataUnavailable	 			if any data needed to build the returned instance of <code>JavaObject</code> is not available.
     * @see 	javax.tools.diagnostics.runtime.java.JavaObject
     */
     JavaObject getObjectAtAddress(ImagePointer address) throws CorruptDataException, IllegalArgumentException, MemoryAccessException, DataUnavailable;
    
	/**
	 * @param obj
	 * @return true if the given object refers to the same Java Runtime in the image
	 */
	public boolean equals(Object obj);
	public int hashCode();

	/**
	 * Returns the File used as source for the creation of this Runtime.
	 * This File will be equal to the file presented to the FactoryRegistry 
	 * when this runtime (or its parent Image) was created.
	 * 
	 * @return File object 
	 */
	File getSource();
}
