/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.jvmti.util;

import java.io.EOFException;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;

import junit.framework.TestCase;

/**
 * Testcases for CachedRandomAccessFileTest.
 * Currently only supports reading. 
 *
 */
public class CachedRandomAccessFileTest extends TestCase {

	CachedRandomAccessFile craf;
	File file;
	
	final static int FILE_SIZE = 187; // Should not be a power of two and must be >127 for tests to work with signed bytes.
	final static int BIG_FILESIZE = 1049576; // about 4MB

	
	protected void setUp() throws Exception {
		super.setUp();
		file = File.createTempFile("CachedRandomAccessFileTest", "testfile");
		RandomAccessFile raf = new RandomAccessFile(file,"rw");
	
		for(int i=0; i < FILE_SIZE; i++) {
			raf.write(i);
		}
		
		raf.close();
		
	}

	protected void tearDown() throws Exception {
		super.tearDown();
		
		file.delete();
	}

	public void testBadFilename() throws Exception {
		try {
			craf = new CachedRandomAccessFile("bad" + file.getName());
			fail("FileNotFoundException not thrown against bad file name");
		} catch(FileNotFoundException e) {
			// Expected
		}
		
		try {
			craf = new CachedRandomAccessFile("bad" + file.getName(), 12, 32);
			fail("FileNotFoundException not thrown against bad file name");
		} catch(FileNotFoundException e) {
			// Expected
		}
	}
	
	public void testBadFile() throws Exception {
		try {
			craf = new CachedRandomAccessFile(new File("bad" + file.getName()));
			fail("FileNotFoundException not thrown against bad file");
		} catch(FileNotFoundException e) {
			// Expected
		}
		
		try {
			craf = new CachedRandomAccessFile(new File("bad" + file.getName()), 12, 32);
			fail("FileNotFoundException not thrown against bad file");
		} catch(FileNotFoundException e) {
			// Expected
		}
		
	}
	
	public void testReadWholeFile() throws Exception {
		craf = new CachedRandomAccessFile(file, 2, 32);
		byte bytes[] = new byte[FILE_SIZE];
		
		craf.read(bytes);
		
		for (int i=0; i < FILE_SIZE-1; i++) {
			assertEquals("Bytes read don't match.", i, bytes[i] & 0xff);
		}
		
		craf.close();
	}
	
	/**
	 * Tests that reading a closed file fails with an IOException.
	 * 
	 * @throws Exception
	 */
	public void testClose() throws Exception {
		craf = new CachedRandomAccessFile(file, 2, 32);
		byte bytes[] = new byte[FILE_SIZE];
		
		craf.read(bytes);
				
		craf.close();

		craf.seek(0);
		
		try{
			craf.read(bytes);
			fail("Failed to throw IOException on closed file.");
		} catch (IOException e) {
			// expected
		}

	}
	/**
	 * Check that when the file is read into an offset
	 * within the array, that only the bytes in the file
	 * are read.
	 * 
	 * @throws Exception
	 */
	public void testReadWholeFileOffset() throws Exception {
		craf = new CachedRandomAccessFile(file, 2, 32);
		byte bytes[] = new byte[256];
		int offset = 27;
		
		craf.read(bytes, offset, FILE_SIZE);
		
		for (int i = 0; i < offset; i++) {
			assertEquals("First portion of array not zeros.", 0, bytes[i]);
		}
		for (int i=0; i < FILE_SIZE-1; i++) {
			assertEquals("Bytes read don't match.", i, bytes[i+offset] & 0xff);
		}
		for (int i = offset+FILE_SIZE; i < bytes.length; i++) {
			assertEquals("End portion of array not zeros.", 0, bytes[i]);
		}
		
		craf.close();
	}
	
	/**
	 * Read a file that is made of a whole block and a 
	 * partial block.
	 * 
	 * @throws Exception
	 */
	public void testReadWholeFilePartialBlock() throws Exception {
		craf = new CachedRandomAccessFile(file, 6, 32);
		byte bytes[] = new byte[FILE_SIZE];
		
		craf.read(bytes);
		
		for (int i=0; i < FILE_SIZE-1; i++) {
			assertEquals("Bytes read don't match.", i, bytes[i] & 0xff);
		}
		
		craf.close();
	}
	
	/**
	 * Check that when the file is read into an offset
	 * within the array, that only the bytes in the file
	 * are read.
	 * 
	 * @throws Exception
	 */
	public void testReadWholeFileOffsetPartialBlock() throws Exception {
		craf = new CachedRandomAccessFile(file, 6, 32);
		byte bytes[] = new byte[256];
		int offset = 27;
		
		craf.read(bytes, offset, FILE_SIZE);
		
		for (int i = 0; i < offset; i++) {
			assertEquals("First portion of array not zeros.", 0, bytes[i]);
		}
		for (int i=0; i < FILE_SIZE-1; i++) {
			assertEquals("Bytes read don't match.", i, bytes[i+offset] & 0xff);
		}
		for (int i = offset+FILE_SIZE; i < bytes.length; i++) {
			assertEquals("End portion of array not zeros.", 0, bytes[i]);
		}
		
		craf.close();
		
	}
	
	/**
	 * Test an uneven seek over the file.
	 * 
	 * @throws Exception
	 */
	public void testSeek() throws Exception {
		craf = new CachedRandomAccessFile(file, 4, 32);
		
		for (int i=0,pos=1; pos < FILE_SIZE; i++, pos=pos+i) {
			craf.seek(pos);
			int value = craf.read();
			
			assertEquals("Byte "+pos+" must be equal", pos, value);
		}
		
		craf.close();
	}
	
	/**
	 * Test a seek beyond the end of the file.
	 * 
	 * @throws Exception
	 */
	public void testSeekBeyondEnd() throws Exception {
		craf = new CachedRandomAccessFile(file, 4, 32);
		int position = FILE_SIZE+200;
		
		craf.seek(position);
		
		assertEquals("New position incorrect.", craf.getFilePointer(), position);
		
		craf.close();
	}
	
	/**
	 * Test a seek beyond the end of the file.
	 * 
	 * @throws Exception
	 */
	public void testSeekBeyondEndAndRead() throws Exception {
		craf = new CachedRandomAccessFile(file, 4, 32);
		int position = FILE_SIZE+200;
		
		craf.seek(position);
		
		
		int value = craf.read();
		
		assertEquals("Read beyond end didn't result in -1.",-1,value);
		
		craf.close();
	}
	
	/**
	 * Test a seek beyond the end of the file.
	 * 
	 * @throws Exception
	 */
	public void testSeekBeyondEndAndReadArray() throws Exception {
		craf = new CachedRandomAccessFile(file, 4, 32);
		int position = FILE_SIZE+200;
		
		craf.seek(position);
		
		byte[] array = {47,47,47,47,47};
		
		int value = craf.read(array);
		
		assertEquals("Read beyond end didn't result in -1.",-1,value);
		
		for (int i=0; i < array.length; i++) {
			assertEquals("Array value was changed.",47,array[i]);
		}
		
		craf.close();
	}
	
	
	/**
	 * Test a seek beyond the end of the file.
	 * 
	 * @throws Exception
	 */
	public void testSeekBeforeBeginning() throws Exception {
		craf = new CachedRandomAccessFile(file, 4, 32);
		int position = -200;
		
		try {
			craf.seek(position);
			fail("Seeking to negative file position failed to throw IOException.");
		} catch (IOException e) {
			// expected
		}
		
		
		craf.close();
	}
	
	/**
	 * IOException should be thrown after a read.
	 * 
	 * 
	 * @throws Exception
	 */
	public void testReadAfterClose() throws Exception {
		craf = new CachedRandomAccessFile(file, 4, 32);
		craf.close();
		
		try{
			craf.seek(64);
			craf.read();
			fail("A read after a close failed to throw an IOException.");
		} catch (IOException e) {
			// Expected
		}
		
	}
	
	
	/**
	 * Test that length reported is correct.
	 * 
	 * @throws Exception
	 */
	public void testLength() throws Exception {
		craf = new CachedRandomAccessFile(file, 4, 32);
		assertEquals("Length is not correct.", FILE_SIZE, craf.length());
		craf.close();		
	}
	
	/**
	 * Test getting the File pointer
	 * 
	 * @throws Exception
	 */
	public void testFilePointer() throws Exception {
		craf = new CachedRandomAccessFile(file, 4, 32);
		
		for (int i=0,pos=1; pos < FILE_SIZE; i++, pos=pos+i) {
			craf.seek(pos);
			assertEquals("File Pointer should be equal", pos,craf.getFilePointer());
			
			int value = craf.read();
			assertEquals("File Pointer should be equal", pos+1,craf.getFilePointer());			
			
			assertEquals("Byte "+pos+" must be equal", pos, value);
		}
		
		craf.seek(0);
		byte[] array = new byte[5];
		craf.read(array);
		
		assertEquals("File pointer incorrect after array read", array.length, craf.getFilePointer());
		
		craf.read(array,1,2);
		assertEquals("File pointer incorrect after array read 2", array.length+2, craf.getFilePointer());
		
		craf.close();
	}	
	
	/**
	 * Test that a read across a boundary works.
	 * 
	 * @throws Exception
	 */
	public void testStraddlingRead() throws Exception {
		craf = new CachedRandomAccessFile(file, 4, 32);
		
		craf.seek(14); 
		
		byte[] array = new byte[4];
		craf.read(array, 0, 4);
		
		for (int i=0; i < 4; i++) {
			assertEquals(i+14, array[i]);
		}
		
		craf.close();
	}
	
	/**
	 * Ensure that repeated reads from cache work ok.
	 * 
	 * @throws Exception
	 */
	public void testRepeatedReads() throws Exception {
		craf = new CachedRandomAccessFile(file, 2, 32);
		
		for (int i=0; i < 12; i++) {
			int value = craf.read();
			
			assertEquals("Index and read value must be equal", i, value);
		}
		
		craf.seek(0);
		
		for (int i=0; i < 12; i++) {
			int value = craf.read();
			
			assertEquals("Index and read value must be equal", i, value);
		}
		craf.seek(0);
		
		for (int i=0; i < 12; i++) {
			int value = craf.read();
			
			assertEquals("Index and read value must be equal", i, value);
		}	
		craf.close();
	}
	
	/**
	 * Test that an integer is read correctly
	 * @throws Exception
	 */
	public void testReadIntLong() throws Exception {
		craf = new CachedRandomAccessFile(file);
		int integer = craf.readInt();
		
		assertEquals("Mismatching integer", 0x00010203, integer);
		integer = craf.readInt();
		
		assertEquals("Mismatching integer", 0x04050607, integer);
		
		long longint = craf.readLong();
		
		assertEquals("Mismatching long", 0x08090a0b0c0d0e0fL, longint);
		
		craf.close();
	}
	
	/**
	 * Tests the ability of file to be read beyond the first block.
	 * 
	 * @throws Exception
	 */
	public void testFarReadLong() throws Exception {
		craf = new CachedRandomAccessFile(file,4,32);
		
		craf.seek(68);
		long value = craf.readLong();
				
		assertEquals("Mismatching long", 0x4445464748494a4bl, value);
		
		craf.close();
	}
	
	/**
	 * Tests that a linear read of the file, byte by byte,
	 * produces a "-1" when the end of the file is reached.
	 * 
	 * @throws Exception
	 */
	public void testReadToEnd() throws Exception {
		craf = new CachedRandomAccessFile(file,4,32);
		
		for(int i =0 ; i < FILE_SIZE+1; i++) {
			int value = craf.read();
			
			if (i == FILE_SIZE) {
				assertEquals("End of file should have caused read() to return -1", -1, value);
			}
		}
		
		craf.close();
	}

	/**
	 * Checks that read being passed null causes an exception to be
	 * thrown.
	 * 
	 * @throws Exception
	 */
	public void testReadArrayNull() throws Exception {
		craf = new CachedRandomAccessFile(file,4,32);
		
		try{
			craf.read(null,0,5);
			fail("read failed to throw NullPointerException.");
		}catch(NullPointerException e) {
			// Expected
		}
		
		craf.close();
	}
	
	/**
	 * Checks that an offset beyond the end of the array is caught.
	 * 
	 * @throws Exception
	 */
	public void testReadOffsetIOB() throws Exception {
		craf = new CachedRandomAccessFile(file,4,32);
		
		byte[] array = new byte[5];
		
		try{
			craf.read(array,array.length,1);
			fail("Read failed to produce IndexOutOfBoundsException.");
		} catch(IndexOutOfBoundsException e) {
			// Expected
		}
		
		// 
		try{
			craf.read(array,array.length+1,1);
			fail("Read failed to produce IndexOutOfBoundsException.");
		} catch(IndexOutOfBoundsException e) {
			// Expected
		}
		craf.close();
	}
	
	
	/**
	 * Tests for IndexOutOfBoundsException when length passed
	 * to read is larger than array size.
	 * 
	 * @throws Exception
	 */
	public void testReadLengthIOB() throws Exception {
		craf = new CachedRandomAccessFile(file,4,32);
		
		byte[] array = new byte[5];
		
		try{
			craf.read(array,0, FILE_SIZE);
			fail("Read failed to produce IndexOutOfBoundsException.");
		} catch(IndexOutOfBoundsException e) {
			// Expected
		}
		
		// 
		try{
			craf.read(array,0,array.length+1);
			fail("Read failed to produce IndexOutOfBoundsException.");
		} catch(IndexOutOfBoundsException e) {
			// Expected
		}
		
		craf.close();
	}
	
	/**
	 * Test for IndexOutOfBoundsException when reading into
	 * array with negative offset.
	 * 
	 * @throws Exception
	 */
	public void testReadNegOffset() throws Exception {
		craf = new CachedRandomAccessFile(file,4,32);
		
		byte[] array = new byte[5];
		
		try{
			craf.read(array,-1, 12);
			fail("Read failed to produce IndexOutOfBoundsException.");
		} catch(IndexOutOfBoundsException e) {
			// Expected
		}
		
		craf.close();
	}
	
	/**
	 * Asserts that an array with zero elements will have
	 * not bytes read.
	 * 
	 * @throws Exception
	 */
	public void testZeroArray() throws Exception {
		craf = new CachedRandomAccessFile(file,4,32);
		
		byte[] array = new byte[0];
		
		int numRead = craf.read(array);
		
		assertEquals("No bytes should be read into zero size array.",0, numRead);
		
		craf.close();
	}
	
	/**
	 * Tests that an array read that straddles the end of
	 * the file can be read.
	 * 
	 * @throws Exception
	 */
	public void testReadStraddleEnd() throws Exception {
		craf = new CachedRandomAccessFile(file,4,32);
		
		byte[] array = new byte[16];
		
		byte init=47;
		
		for (int i=0; i< array.length; i++) {
			array[i] = init;
		}
		
		int offset = 8;
		craf.seek(FILE_SIZE-offset);
		
		int num=craf.read(array);
		
		assertEquals("End of file was reached, but bytes read was not returned.", offset,num);
		
		// Check that the bytes that should have been read are read.
		for(int i=0; i<offset; i++) {
			assertEquals("Bytes read did not match",FILE_SIZE-offset+i, array[i] & 0xff);
		}
		
		// check that the part of the array that could not be read is
		// left as-is.
		for(int i=offset; i<array.length; i++) {
			assertEquals("Unread section of array have been written.",init,array[i]);
		}
		
		craf.close();
	}
	
	/**
	 * Test the skip bytes method.
	 * 
	 * @throws Exception
	 */
	public void testSkipBytes() throws Exception {
		craf = new CachedRandomAccessFile(file, 4, 1);
		
		int value;
		
		value = craf.read();
		
		assertEquals("Wrong byte was read.",0, value);
		
		int res = craf.skipBytes(10);
		
		assertEquals(10,res);

		value = craf.read();
		
		assertEquals("Wrong byte was read.",11, value);

		res = craf.skipBytes(10);
		
		assertEquals(10,res);

		value = craf.read();

		craf.close();
	}
	
	/**
	 * Check that skipping over 0 bytes has expected behaviour.
	 * 
	 * @throws Exception
	 */
	public void test0SkipBytes() throws Exception {
		craf = new CachedRandomAccessFile(file, 4, 1);
		
		craf.seek(23);
		
		int value;
		
		value = craf.read();
		
		assertEquals("First read value incorrect.", 23, value);
		
		int num = craf.skipBytes(0);
		
		assertEquals("skipBytes(0) returned non-zero",0,num);
		
		value = craf.read();
		
		assertEquals("Second read value incorrect.", 24, value);
		
		craf.close();
	}
	
	/**
	 * Check that skipping over -1 bytes has expected behaviour.
	 * 
	 * @throws Exception
	 */
	public void test_1SkipBytes() throws Exception {
		craf = new CachedRandomAccessFile(file, 4, 1);
		
		craf.seek(23);
		
		int value;
		
		value = craf.read();
		
		assertEquals("First read value incorrect.", 23, value);
		
		int num = craf.skipBytes(-1);
		
		assertEquals("skipBytes(-1) returned non-zero",0,num);
		
		value = craf.read();
		
		assertEquals("Second read value incorrect.", 24, value);
		
		assertEquals("File position not correct.", craf.getFilePointer(), value+1);
		
		craf.close();
	}
	
	/**
	 * Check that skipBytes() will skip only as far 
	 * 
	 * @throws Exception
	 */
	public void testSkipBytesTOEOF() throws Exception {
		craf = new CachedRandomAccessFile(file, 4, 1);
		int offset = 9;
		
		craf.seek(FILE_SIZE-offset);
		
		int num;
		
		num = craf.skipBytes(offset+20);
		
		assertEquals(offset,num);
		
		craf.close();
	}
	
	/**
	 * Test a big file that will overflow the cache.
	 * Previous tests won't cause cache  
	 * 
	 * @throws Exception
	 */
	public void testBigFileSmallCache() throws Exception {		
		File bigfile = File.createTempFile("CachedRandomAccessFileTest", "testfile");
		RandomAccessFile raf = new RandomAccessFile(bigfile,"rw");
	
		for(int i=0; i < BIG_FILESIZE; i++) {
			raf.writeInt(i);
		}		
		
		raf.close();
	
		craf = new CachedRandomAccessFile(bigfile, 12,1);
		
		for (int i = 0; i < BIG_FILESIZE; i++) {
			int value = craf.readInt();
			
			assertEquals("Int values did not match",i, value);
		}
		
		craf.close();
		
		bigfile.delete();
	}
	
	/**
	 * Test that reading a long will produce an EOFException if it occurs
	 * at the end of the file.
	 * 
	 * @throws Exception
	 */
	public void testReadLongEOF() throws Exception {
		craf = new CachedRandomAccessFile(file, 4, 1);
	
		craf.seek(FILE_SIZE-2);
		
		try {
			long value = craf.readLong();
			fail("Expected EOFException not thrown by readLong. Got 0x"+Long.toHexString(value));
		} catch (EOFException e) {
			// Expected
		}
		
		craf.close();
	}
	
	
	/* 
	 * The following tests test for IOExceptions from the methods that modify the file.	
	 * In principle the CachedRandomAccessFileTest could have write capabilities.
	 */
	
	/**
	 * Ensures write method throws IOException. 
	 * 
	 * @throw Exception
	 */
	public void testWrite() throws Exception {
		craf = new CachedRandomAccessFile(file);
		
		try{
			craf.write(10);
			fail("write(int b) failed to throw IOException.");
		} catch (IOException e) {
			// Expected
		}
		
		craf.close();
	}
	
	/**
	 * Ensures write method throws IOException 
	 * 
	 * @throw Exception
	 */
	public void testWriteArray() throws Exception {
		craf = new CachedRandomAccessFile(file);
		
		byte[] array = {10, 20, 30};
		try{
			craf.write(array);
			fail("write(byte[] b) failed to throw IOException.");
		} catch (IOException e) {
			// Expected
		}
		
		craf.close();
	}
	
	/**
	 * Ensures write method throws IOException 
	 * 
	 * @throw Exception
	 */
	public void testWriteArrayOffsets() throws Exception {
		craf = new CachedRandomAccessFile(file);
		
		byte[] array = {10, 20, 30};
		try{
			craf.write(array, 0, array.length);
			fail("write(byte[], int off, int len) failed to throw IOException.");
		} catch (IOException e) {
			// Expected
		}
		
		craf.close();
	}
	
	/**
	 * Ensure setLength fails with IO Exception.
	 * @throws Exception
	 */
	public void testSetLength() throws Exception {
		craf = new CachedRandomAccessFile(file);
		
		try {
			craf.setLength(10);
			fail("setLength(long length) failed to throw IOException.");
		} catch (IOException e) {
			// Expected.
		}
		
		craf.close();
	}
	
	
}
