/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.hprof.java;

import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.runtime.java.JavaClassLoader;
import javax.tools.diagnostics.runtime.java.JavaObject;

import org.apache.kato.common.IteratorBackedList;
import org.apache.kato.hprof.IJavaClassLoader;
import org.apache.kato.hprof.image.CorruptDataImpl;

/**
 * Implements JavaClassLoader.
 * 
 *
 */
public class JavaClassLoaderImpl implements JavaClassLoader {
	IJavaClassLoader loader;
	IJavaHeapInternal heap;
	
	public JavaClassLoaderImpl(IJavaHeapInternal heap, IJavaClassLoader loader) {
		this.loader = loader;
		this.heap = heap;;
	}
	
	/**
	 * Returns the JavaClassImpl that has the name passed.
	 * 
	 * Iterates through all of the Class object IDs, and retrieves
	 * the IJavaClasses held in HProfView, so as not to create JavaClasses
	 * unnecessarily.
	 */
	@Override
	public JavaClassImpl findClass(String className) throws CorruptDataException {
		if (className == null) {
			return null;
		} else if (className.length() == 0) {
			return null;
		}
		
		Collection<Long> classes = loader.getClasses();
		
		for (Long ID : classes) {
			JavaClassImpl clazz = heap.getJavaClassByID(ID);
			if (className.equals(clazz.getName())) {
				return clazz;
			}
		}
		
		return null;
	}

	@Override
	/**
	 * Returns an empty Iterator, as the HProf format doesn't describe
	 * how the classes are cached.
	 * 
	 * FIXME It may be possible to pull out the classes from the java.lang.ClassLoader.classes Vector 
	 */
	public List getCachedClasses() {
		return new LinkedList();
	}

	@Override
	public List getDefinedClasses() {
		return new IteratorBackedList(new ClassesIterator());
	}

	/**
	 * An iterator over the classes loaded by this class loader.
	 * Wraps the IDs this class loader holds with JavaClassImpls.
	 * 
	 * @author monteith
	 *
	 */
	private class ClassesIterator implements Iterator {
		private Collection<Long> classes = loader.getClasses();
		private Iterator<Long> iterator = classes.iterator();		
		@Override
		public boolean hasNext() {
			return iterator.hasNext();
		}

		@Override
		public Object next() {
			Long ID = iterator.next();
			JavaClassImpl clazz = heap.getJavaClassByID(ID);

			if (clazz == null) {
				return new CorruptDataImpl(heap.getImagePointer(ID), "Can't match class object ID to a JavaClassImpl.");
			}
			
			return clazz;
		}

		/**
		 * Not implemented.
		 */
		@Override
		public void remove() {
			throw new UnsupportedOperationException("remove() not supported for ClassesIterator");
		}
		
	}
	
	/**
	 * Gets the java/lang/ClassLoader object from the heap.
	 * 
	 */
	@Override
	public JavaObject getObject() throws CorruptDataException {
		return heap.getObjectByID(this.loader.getID());
	}

}
