/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.hprof.java;

import java.io.File;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.ImageAddressSpace;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.image.MemoryAccessException;
import javax.tools.diagnostics.runtime.java.JavaHeap;
import javax.tools.diagnostics.runtime.java.JavaObject;
import javax.tools.diagnostics.runtime.java.JavaRuntime;
import javax.tools.diagnostics.runtime.java.JavaVMInitArgs;

import org.apache.kato.hprof.HProfView;
import org.apache.kato.hprof.IJavaThread;
import org.apache.kato.hprof.image.CorruptDataImpl;
import org.apache.kato.hprof.image.SimpleImagePointer;

public class JavaRuntimeImpl implements JavaRuntime {

	private HProfView view=null;
	public JavaRuntimeImpl(HProfView view, ImageAddressSpace addressSpace) {
		this.view = view;
		heap = new JavaHeapImpl(view, view.getHeapRecordLocation(), addressSpace);
	}

	/**
	 * Return the HProfView this JavaRuntime is based on.
	 * 
	 * @return HProfView.
	 */
	HProfView getView() {
		return view;
	}
	
	@Override
	public List getCompiledMethods() {
		
		
		List l=new LinkedList();
		return l;
	}

	@Override
	public List getHeapRoots() {
		List l=new LinkedList();
		return l;
	}

	private JavaHeapImpl heap;
	
	@Override
	public List getHeaps() {
		List<JavaHeapImpl> l=new LinkedList<JavaHeapImpl>();
		l.add(heap);
		return l;
	}

	
	
	/**
	 * Retrieves all of the JavaClassLoaders owned by this JavaRuntime.
	 * 
	 * 
	 * @return Iterator<JavaClassLoaderImpl>
	 */
	@Override	
	public List getJavaClassLoaders() {
		return heap.getJavaClassLoaders();
	}

	
	@Override
	public ImagePointer getJavaVM() throws CorruptDataException {
		
		return new SimpleImagePointer(0);
	}

	@Override
	public JavaVMInitArgs getJavaVMInitArgs() throws DataUnavailable,
			CorruptDataException {
			throw new DataUnavailable();
			
	}

	/**
	 * 
	 * Monitors aren't present in HProf, so return an empty iterator.
	 * 
	 * @return empty Iterator
	 */
	@Override
	public List getMonitors() {		
		return new LinkedList();
	}

	/**
	 * Returns an object with a given object ID.
	 * This is not the same as the objects address, which is unknown in HProf.
	 * This method is the mirror of JavaObject.getID().
	 * 
	 * @return JavaObject with the given object ID.
	 * @throws IllegalArgumentException If an object with the given ID can't be found.
	 * 
	 */
	@Override
	public JavaObject getObjectAtAddress(ImagePointer address)
			throws CorruptDataException, IllegalArgumentException,
			MemoryAccessException, DataUnavailable {
		
		if(address==null) {
			throw new IllegalArgumentException("address is null");	
		}
		JavaObject obj = heap.getObjectByID(address.getAddress());
		
		if (obj == null) {
			throw new IllegalArgumentException("Couldn't retrieve object @"+
					Long.toHexString(address.getAddress()));
		}
		
		return obj;
	}

	private List<JavaThreadImpl> threads = null;
	@Override
	public List getThreads() {
		if(threads == null) {
			threads = new LinkedList<JavaThreadImpl>();

			Collection<IJavaThread> ithreads = view.getActiveThreads();
			Iterator<IJavaThread> iter = ithreads.iterator();
			
			while (iter.hasNext()) {
				
				threads.add(new JavaThreadImpl(heap, iter.next()));
			}		
		}
		
		return threads;
	}

	/**
	 * HProf could potentially return trace buffers, through analysing the 
	 * stack trace records, etc. 
	 * This isn't done now - the emphasis is on the heap, and threads.
	 * This method may even be removed.
	 */
	@Override
	public Object getTraceBuffer(String bufferName, boolean formatted)
			throws CorruptDataException {
		throw new CorruptDataException(new CorruptDataImpl("There are no trace buffers"));
	}

	@Override
	public String getFullVersion() throws CorruptDataException {
		// TODO Auto-generated method stub
		return "Hprof dump.";
	}

	@Override
	public String getVersion() throws CorruptDataException {
		// TODO Auto-generated method stub
		return "Hprof dump.";
	}
	


	JavaHeap getHeap() {
		return heap;
	}

	@Override
	public File getSource() {
		
		return view.getSource();
	}
}
