/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.hprof.java;

import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.image.ImageThread;
import javax.tools.diagnostics.runtime.java.JavaObject;
import javax.tools.diagnostics.runtime.java.JavaThread;

import org.apache.kato.hprof.IJavaStack;
import org.apache.kato.hprof.IJavaStackFrame;
import org.apache.kato.hprof.IJavaThread;

/**
 * Implements a JavaThread by wrapping a IJavaThread from HprofView.
 *
 */
public class JavaThreadImpl implements JavaThread {
	private IJavaHeapInternal heap;
	private IJavaThread thread;
	
	public JavaThreadImpl(IJavaHeapInternal heap, IJavaThread thread) {
		this.heap = heap;
		this.thread = thread;
	}
	
	/**
	 * There are no image threads in hprof files.
	 * 
	 * @throws DataUnavailable always - there are no ImageThreads
	 */
	@Override
	public ImageThread getImageThread() throws CorruptDataException,
			DataUnavailable {
		throw new DataUnavailable("ImageThreads not present in Hprof file.");
	}

	
	/**
	 * There are no JNIEnvs in Hprof.
	 * There is no DataUnavailable exception, so we wrap up the thread serial# in an
	 * image pointer.
	 * 
	 * @return ImagePointer of thread serial number
	 */
	@Override
	public ImagePointer getJNIEnv() throws CorruptDataException {
		return heap.getImagePointer(thread.getThreadSerialNumber());
	}

	@Override
	public String getName() throws CorruptDataException {
		return thread.getThreadName();
	}

	@Override
	public JavaObject getObject() throws CorruptDataException {
		return heap.getObjectByID(thread.getThreadObjectID());
	}

	/**
	 * Priority not stored in Hprof.
	 * 
	 * @return 5
	 */
	@Override
	public int getPriority() throws CorruptDataException {		
		return 5;
	}

	/**
	 * 
	 * @return Iterator over all stack frames
	 */
	private Vector<JavaStackFrameImpl> frames;
	@Override
	public List getStackFrames() {
		if(frames == null) {
			IJavaStack stack =thread.getStack();
		 	IJavaStackFrame[] stackFrames = stack.getStack();
		 	frames = new Vector<JavaStackFrameImpl>(stackFrames.length);

		 	for (IJavaStackFrame frame : stackFrames) {
		 		frames.add(new JavaStackFrameImpl(heap, frame));
		 	}
		}
		
		return frames;
	}

	/**
	 * Stack sections are not recorded in hprof files.
	 * 
	 * @return empty iterator.
	 */
	@Override
	public List getStackSections() {
		return new LinkedList();
	}

	/**
	 * State isn't recorded in Hprof. Only live threads
	 * are reported, STATE_ALIVE can therefore be stated.
	 * 
	 * @return STATE_ALIVE
	 */
	@Override
	public int getState() throws CorruptDataException {
		return STATE_ALIVE;
	}

}
