/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.hprof.datalayer;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.List;

import javax.imageio.stream.FileImageInputStream;

import org.apache.kato.common.IDataProvider;

public class RandomAccesDataProvider implements IDataProvider {

	private FileImageInputStream raf=null;
	private long fileLength=0;
	private long current=0;
	
	private boolean moved=false;
	
	public RandomAccesDataProvider(File reader) throws IOException {
		if(reader==null) throw new IllegalArgumentException("reader file is null");
		if(reader.exists()==false) throw new IllegalArgumentException("reader file ["+reader.getAbsolutePath()+ " ] does not exist");
		RandomAccessFile r=new RandomAccessFile(reader, "r");
		raf=new FileImageInputStream(r);
	}

	@Override
	public void close() throws IOException {
		raf.close();
		raf=null;
		
	}

	@Override
	public void moveTo(long l) throws IOException {
		
		current=l;
		moved=true;
		
	}

	@Override
	public void open() throws IOException {
		fileLength=raf.length();
		raf.seek(0);
		
	}

	@Override
	public String readCString() throws IOException {
		updateLocation();
		
		StringBuilder builder=new StringBuilder();
		while(true) {
			int b=raf.readUnsignedByte();
			current++;
			if(b==0) break;
			builder.append( (char)b);
		}
		
		return builder.toString();
	}

	@Override
	public int readU4() throws IOException {
		updateLocation();
		current+=4;
		return raf.readInt();
	}

	@Override
	public long readU8() throws IOException {
		updateLocation();
		current+=8;
		long high=raf.readUnsignedInt();
		long low=raf.readUnsignedInt();
		return high << 32 | low;
	}

	@Override
	public long getCurrentLocation() throws IOException {
		
		return current;
	}

	@Override
	public long getDataLength() throws IOException {
		
		return fileLength;
	}

	@Override
	public short readByte() throws IOException {
		updateLocation();
		current++;
		return (short) (raf.readByte() & 0xff);
	}

	@Override
	public byte[] readBytes(int left) throws IOException {
		updateLocation();
		current+=left;
		byte[] data=new byte[left];
		 raf.readFully(data);
		return data; 
	}

	@Override
	public void moveBy(int move) throws IOException {
		current+=move;
		moved=true;
		
	}

	@Override
	public short readU2() throws IOException {
		updateLocation();
		current+=2;
		return raf.readShort();
	}
	private void updateLocation() throws IOException {
		if(moved) {
			raf.seek(current);
			moved=false;
		}
		
	}

	@Override
	public void addState(List state) {
		long realPos=-2;
		try {
			realPos = raf.getStreamPosition();
		} catch (IOException e) {
		
		}
		
		state.add("RandomAccessDataProvider current="+current+",moved="+moved+",fileLength="+fileLength+",realpos="+realPos);
		
	}

	@Override
	public long dataRemaining() {
		
		return fileLength-current;
	}
	
	@Override
	public boolean moved() {
		return moved;
	}
}
