/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tck.api;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.TreeSet;


public class PropertyStore {

	private PropertyStore parent=null;
	private Properties current=null;
	private String source=null;
	public String getSource() {
		return source;
	}

	private boolean debug=false;
	
	public PropertyStore() {
		this(new Properties(),"no source");
		
	}
	public PropertyStore(Properties current,String source,PropertyStore parent) {
		this(current,source);
		this.parent=parent;
	}
	public PropertyStore( Properties current,String source) {
		this.current=current;
		this.source=source;
	}
	
	
	public PropertyStore(PropertyStore parent) {
		this();
		this.parent=parent;
	}
	public PropertyStore(File propertiesFile) throws IOException {
		
		current=new Properties();
		source=propertiesFile.getAbsolutePath();
		FileInputStream fis = new FileInputStream(propertiesFile);
		current.load(fis);
		fis.close();
		
		
		
	}
	
	public PropertyStore(InputStream propertyStream,String source) throws IOException {
		current=new Properties();
		current.load(propertyStream);
		this.source=source;
		
		
	}
	public PropertyStore(InputStream propertyStream,String source,PropertyStore parent) throws IOException {
		current=new Properties();
		current.load(propertyStream);
		this.parent=parent;
		this.source=source;
	}
	
	public PropertyStore(File localFile, PropertyStore parent) throws IOException {
		this(localFile);
		this.source=localFile.getAbsolutePath();
		this.parent=parent;
	}
	public String getProperty(String keyName) {
		String value=retrieveProperty(keyName);
		if(value!=null) {
			value=expandValue(value);
		}
		return value;
	}
	public String getPropertyAsIs(String keyName) {
		
		return current.getProperty(keyName);
	}
	
	private String retrieveProperty(String keyName) {
		
		String value=null;
		if(current.containsKey(keyName)) {
			value=current.getProperty(keyName);
		}
		else {
			if(parent!=null) value=parent.retrieveProperty(keyName);
		}
		
		return value;
	}
	
	
	public  String expandValue(String input) {
		
		 String before=null;
		do {
		  before=input;
		 
		 input=expandString( before);
		 if(debug) {
			 System.out.println(">"+before);
			 System.out.println("<"+input);
		 }
		} while (input.equals(before)==false);
	
		return input;
	}
	
	private String expandString(String input) {
		
		int lastPos=0;
		int suffixPos=0;
		int prefixPos=0;
		while(true) {
			
			suffixPos=input.indexOf("}",suffixPos);
			if(suffixPos<0) {
				// no more suffixes
				return input;
			}
			prefixPos=input.lastIndexOf("${", suffixPos);
			if(prefixPos<0) {
				// not found. move to next suffix
				continue;
			}
			 
			// found a prefix
			// find the matching suffix
			
			int endSuffix=input.indexOf("}",prefixPos);
			break;
		}

		// now we have the works
		
			String propertyName=input.substring(prefixPos+2,suffixPos);
			String replaceValue=null;
			if(hasProperty(propertyName)) replaceValue=retrieveProperty(propertyName);
			
			if(replaceValue==null) replaceValue="?"+propertyName+"?";  
			
			// replace value in string
			
			String left="";
			if(prefixPos>0) left=input.substring(0,prefixPos);
		
			String right="";
			if(suffixPos<input.length()) right=input.substring(suffixPos+1);
			
			input=left+replaceValue+right;
			lastPos=left.length()+replaceValue.length();
			
				
		
		return input;
	}
	public boolean hasProperty(String keyName) {
		
		if(current.containsKey(keyName)) return true;
		if(parent==null)  return false;
		return parent.hasProperty(keyName);
	}
	public void setProperty(String key, String value) {
		current.setProperty(key,value);
		
	}
	/**
	 * Returns a property as a String array
	 * The property value is split in occurances of the seperator
	 * Any replaceable values are resolved prior to splitting 
	 * the value.
	 * 
	 * @param keyname
	 * @param seperator
	 * 
	 */
	public String[] getPropertyAsArray(String keyName, String seperator) {
		
		String value=getProperty(keyName);
		if(value==null) return null;
		
		List list=new LinkedList();
		int sepPos=value.indexOf(seperator);
		if(sepPos<0) {
			list.add(value);
		}
		else {
			while(true) {
			String portion=value.substring(0,sepPos);
			list.add(portion);
			value=value.substring(sepPos+seperator.length());
			sepPos=value.indexOf(seperator);
			if(sepPos<0) {
				list.add(value);
				break;
			}
			}
		
		}
		
		return (String[]) list.toArray(new String[0]);
	}
	/**
	 * Returns an iterator of all keys in the property store chain
	 * 
	 */
	public Iterator iterator() {
		
		Set keys=new HashSet();
		retrieveKeys(keys);
		return keys.iterator();
		
		
	}
	public void retrieveKeys(Set keys) {
		
		keys.addAll(current.keySet());
		if(parent!=null) parent.retrieveKeys(keys);
		
	}
	/**
	 * Returns all property Keys that start with the provided prefix
	 * @param string
	 * 
	 */
	public String[] getPropertyNames(String prefix) {
		
		Set keys=new TreeSet();
		retrieveKeys(keys);
		
		List results=new LinkedList();
		
		Iterator keyIterator=keys.iterator();
		while(keyIterator.hasNext()) {
			String key=(String) keyIterator.next();
			if(key.startsWith(prefix)) results.add(key);
		}
		
		return (String[]) results.toArray(new String[0]); 
	}
	
	public String getRequiredProperty(String key) {
		String result=getProperty(key);
		if(result==null) throw new TCKException("missing property "+key);
		return result;
	}
	public String getProperty(String key, String defaultValue) {
		String result=getProperty(key);
		if(result==null) result=defaultValue;
		return result;
	}
	
	public boolean getPropertyAsBoolean(String key, boolean defaultValue) {

		String value=getProperty(key);
		if(value==null || value.trim().equals("")) return defaultValue;

		value=value.toLowerCase().trim();
		if(value.equals("true")) return true;
		return false;

	}
	public PropertyStore getParent() {
		return parent;
	}
}
