/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

package org.apache.kato.tck.harness.scenario;

import java.io.IOException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.apache.kato.tck.api.ICheckpoint;
import org.apache.kato.tck.api.TCKException;


/**
 * 
 */
public class ScenarioSetupHandler implements IScenarioSetupHandler {

	private CheckPointHandler checkpointhandler = new CheckPointHandler();
	private List scenarioElements = new LinkedList();
	private IDumpTrigger trigger=null;
	
	public ScenarioSetupHandler(IDumpTrigger trigger) {
		this.trigger=trigger;
	}

	/**
	 * Configure an instance of IScenarioElement
	 * All declared methods are scanned and those that are
	 * prefixed with "config" and have no return value , are public
	 * and not static are considered.
	 * If the method takes no parameters or its take a single parameter
	 * of type ICheckpoint then it is called otherwise it is ignored
	 *
	 */
	public void execute(ScenarioElement element) {
		
		// record now to anchor the object 
		// run configure when all have been registered
		
		scenarioElements.add(element);
		
				
	}

	/* (non-Javadoc)
	 * @see org.apache.kato.tck.harness.scenario.IScenerioSetupHandler#close()
	 */
	public void close() throws IOException {
		
		Iterator i=scenarioElements.iterator();
		
		while(i.hasNext()) {
			
			ScenarioElement element=(ScenarioElement) i.next();
			
			configureElement(element);
		}
		
		while(checkpointhandler.hasOutstandingThreads()) {
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
				
				e.printStackTrace();
			}
		}
		trigger.dump();
	}

	/* (non-Javadoc)
	 * @see org.apache.kato.tck.harness.scenario.IScenerioSetupHandler#open()
	 */
	public void open() throws IOException {
		checkpointhandler=new CheckPointHandler();
		scenarioElements=new LinkedList();
	}

	/* (non-Javadoc)
	 * @see org.apache.kato.tck.harness.scenario.IScenerioSetupHandler#getStatus()
	 */
	public String getStatus() {
		
		return "Scenerio Setup : elements registered="+scenarioElements.size()+" threads started="+checkpointhandler.getThreadCount();
	}
	
	private void configureElement(ScenarioElement element) {
		
		
		// get methods
		Class me = element.getClass();
		Method m[] = me.getDeclaredMethods();
	
		// run all methods where the method name starts with configure
		for (int j = 0; j < m.length; j++) {
			Method method = m[j];
			String methodName = method.getName();
			int mods=method.getModifiers();
			
			if (methodName.startsWith("config")==false) continue; 
		
			
			// ignore non public and methods that return values
			if(Modifier.isPublic(mods) ==false) continue;
			if(Modifier.isStatic(mods) ==true ) continue;
			if(method.getReturnType().equals(void.class)==false) continue;
			
				
				try {
					Class[] parms = method.getParameterTypes();
					if (parms == null || parms.length == 0) {
						method.invoke(element, null);
					} else {
						if (parms[0].isAssignableFrom(ICheckpoint.class)) {
							
							checkpointhandler.handle(method, element);
						} 
						
					}
	
				} catch (Exception e) {
					e.printStackTrace();
					throw new TCKException("Exception using configure method "
							+ method, e);
	
				}
			}

	}

}
