/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tests.junit;

import java.util.Iterator;
import java.util.List;

import javax.tools.diagnostics.image.CorruptData;
import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.Image;


/**
 * 
 * A collection of simple test-cases to make sure that the DTFJ Image Interface agrees with its documented
 * behaviour.
 */
public class ImageTest extends AbstractImageTestcase
{
	private Image _core=null;
	protected void setUp() throws Exception
	{
		try {
			//this is required to load the core file image which is to be tested so that we don't have to load each time
			_core = defaultTestingImage();
		} catch (Throwable t) {
			t.printStackTrace();
			fail();
		}
		super.setUp();
	}
	
	
	//now, onto the testing
	
	/**
	 * ensure that the core has non-null address spaces and at least one
	 */
	public void testGetAddressSpaces() {
		List it = _core.getAddressSpaces();
		
		assertNotNull(it);
		assertFalse(it.isEmpty());
	}
	
	/**
	 * make sure that the creation time is sane (>0)
	 */
	public void testGetCreationTime() {
		try {
			assertTrue(_core.getCreationTime() > 0);
		} catch (DataUnavailable e) {
			//spec correct
		} catch (Throwable t) {
			//exception fall-through
			fail();
		}
	}
	
	/**
	 * test that the installed memory is sane or not known
	 */
	public void testGetInstalledMemory() {
		try {
			assertTrue(_core.getInstalledMemory() > 0);
		} catch (DataUnavailable e) {
			//this is considered acceptable behaviour
		} catch (Throwable t) {
			//exception fall-through
			fail();
		}
	}
	
	/**
	 * make sure that we get a sane number of processors or unknown
	 */
	public void testGetProcessorCount() {
		try {
			assertTrue(_core.getProcessorCount() > 0);
		} catch (DataUnavailable e) {
			//this is considered acceptable behaviour
		} catch (Throwable t) {
			//exception fall-through
			fail();
		}
	}
	
	/**
	 * Make sure that the processor subtype is non-null
	 * TODO:  this probably isn't the right way to test this
	 */
	public void testGetProcessorSubtype() {
		try {
			assertNotNull(_core.getProcessorSubType());
		} catch (DataUnavailable e) {
			//accepted in spec
		} catch (CorruptDataException e) {
			//accepted in spec
		} catch (Throwable t) {
			//exception fall-through
			fail();
		}
	}
	
	/**
	 * Make sure that we get a processor name
	 */
	public void testGetProcessorType() {
		try {
			assertNotNull(_core.getProcessorType());
		} catch (DataUnavailable e) {
			//this exception is supported by the API so it is ok
		} catch (CorruptDataException e) {
			//this exception is supported by the API so it is ok
		} catch (Throwable t) {
			//exception fall-through
			fail();
		}
	}
	
	/**
	 * Make sure that there is a system sub-type
	 * TODO:  this probably isn't the right way to test this
	 */
	public void testGetSystemSubType() {
		try {
			assertNotNull(_core.getSystemSubType());
		} catch (DataUnavailable e) {
			//accepted in spec
		} catch (CorruptDataException e) {
			//accepted in spec
		} catch (Throwable t) {
			//exception fall-through
			fail();
		}
	}
	
	/**
	 * Make sure that there is a system type
	 */
	public void testGetSystemType() {
		try {
			assertNotNull(_core.getSystemType());
		} catch (DataUnavailable e) {
			//this exception is supported by the API so it is ok
		} catch (CorruptDataException e) {
			//this exception is supported by the API so it is ok
		} catch (Throwable t) {
			//exception fall-through
			fail();
		}
	}
	
	/**
	 * Ensures that the string is non-null and throws no unexpected exceptions
	 */
	public void testGetHostName()
	{
		try {
			assertNotNull(_core.getHostName());
		} catch (DataUnavailable e) {
			//this exception is supported by the API so it is ok
		} catch (CorruptDataException e) {
			//this exception is supported by the API so it is ok
		} catch (Throwable t) {
			//exception fall-through
			fail();
		}
	}
	
	/**
	 * Ensures that the iterator is non-null, throws no expected exceptions, and every member is an InetAddress
	 */
	public void testGetIPAddresses()
	{
		try {
			Iterator addresses = _core.getIPAddresses();
			assertNotNull(addresses);
			while (addresses.hasNext()) {
				Object address = addresses.next();
				assertTrue((address instanceof java.net.InetAddress) 
						|| (address instanceof CorruptData));
			}
		} catch (DataUnavailable e) {
			//this exception is supported by the API so it is ok
		} catch (Throwable t) {
			//exception fall-through
			fail();
		}
	}
}
