/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Stack;

import javax.tools.diagnostics.FactoryRegistry;
import javax.tools.diagnostics.image.Image;
import javax.tools.diagnostics.image.ImageAddressSpace;
import javax.tools.diagnostics.image.ImageFactory;

import org.apache.kato.katoview.commands.RootCommand;
import org.apache.kato.katoview.commands.helpers.Utils;


public class Session {

	public static final String DUMP_FILE_PATH_PROPERTY = "dump_file_path";
	private Image loadedImage;
	private RootCommand rootCommand;
	private HashMap variables;
	private Output out;
	
	private static final String defaultFactoryName = ""; //TODO: assign proper factory class name 
	private static final String factoryProperty = "javax.tools.diagnostics.image.factoryclass";
	private static final String launcherProperty = "javax.tools.diagnostics.tools.katoview.launcher";
	
	public Session(String[] args) {
		sessionInit(args);
	}

	private void sessionInit(String[] args){
		variables = new HashMap();
		out = new Output();

		out.addChannel(new ConsoleOutputChannel());
		
		rootCommand = new RootCommand(out);
		/*add the root_command object to HashMap so that 
		commands such as + and - can access it to invoke hexdump command*/
		variables.put(Utils.RootCommand_OBJECT, rootCommand);
		variables.put("pwd", new File(System.getProperty("user.dir")));
		
		loadedImage = imageFromCommandLine(args);
	}
	
	/**
	 * @param args
	 */
	private Image imageFromCommandLine(String args[]) {
		Image loadedImage = null;
		int dumpFilePathIndex = indexOf("-dump", args)+1;
		int verboseIndex = indexOf("-verbose", args)+1;		
		
		if (0 != verboseIndex) {
			variables.put("verbose.mode", "on");
		}
		
		String launcher = System.getProperty(launcherProperty, "katoview");
		
		// User must specify -core as parameter, and a dump filename following
		if (((0 == dumpFilePathIndex) && (args.length <= dumpFilePathIndex))) {
			out.println(
				"\n\tUsage: \"" + launcher + " [ -verbose ] -dump <dump_file>");
			return null;
		}

//		out.println(Version.getAllVersionInfo(factory));
		out.println("Loading dump from Kato...\n");
		
		try {		    
			String coreFilePath = args[dumpFilePathIndex];
			loadedImage = FactoryRegistry.getDefaultRegistry().getImage(new File(coreFilePath));
			
			if (loadedImage == null) {
				throw new IOException("No ImageFactory could load "+coreFilePath);
			}
			
			//Store the core file file path in the variables. We need it for creating default filenames for heapdumps
			variables.put(DUMP_FILE_PATH_PROPERTY, coreFilePath);
		} catch (IOException e) {
			out.error("Could not load dump file: " + e.getMessage());
			
			if (null != variables.get("verbose.mode")) {
				e.printStackTrace();
			}
		}
		
		return loadedImage;
	}

	/**
	 * @param string, args
	 */
	private int indexOf(String string, String[] args) {
		int index = -1;
		
		for (int x = 0; x < args.length; x++) {
			if (string.equals(args[x])) {
				index = x;
				break;
			}
		}
		return index;
	}
	
	public void run(){
		BufferedReader stdin = new BufferedReader(new InputStreamReader(System.in));
		String input = "";
		String quit = "";
		String prompt = "> ";
		Stack inputWordsStack = new Stack();
		
		if (null == loadedImage)
			return;
		
		// FIXME: In the current code, this is the _only_ place where the current
		//  address space is set.  Also, the only commands that use this value (at the
		//  time of this writing, anyway) are the "x/k" command (XKCommand class) and
		//  the "info thread" command.  What we would ideally like is some
		//  sort of command that sets the current address space based on some sort of
		//  identifier (there is currently no identifier; that would have to be built
		//  into the Kato API).  Also, we would need all commands (at least those that
		//  are dependent on accessing memory in an address space) to use this value;
		//  currently only "x/k" and "info thread" use this, as stated above.  We don't
		//  really know if multiple address spaces can actually exist in a given core
		//  file, but this capability is put in here just in case they can.
		
		variables.put(
				"current_address_space",
				(ImageAddressSpace)loadedImage.getAddressSpaces().get(0)
		);
		
		out.println("For a list of commands, type \"help\"; " +
				"for how to use \"help\", type \"help help\"");

		while (quit == null || !quit.equals("on"))
		{
			out.printPrompt(prompt);
			
			try {
				input = stdin.readLine();
			} catch (IOException e) {
				out.error("IOException encountered while reading input; exiting program...");
				break;
			}
			
			if (null == input)
			{
				out.error("End of input stream has been reached; exiting program...");
				break;
			}
			
			String trimmedInput = input.trim();
			out.printInput(System.currentTimeMillis(), prompt, input);
			
			if (!trimmedInput.equals(""))
			{
				inputWordsStack = Utils.constructStackFromString(trimmedInput);
				try {
					rootCommand.doCommand(inputWordsStack, loadedImage, variables);
				} catch (RuntimeException e) {
					out.error("An error occurred while processing the command: " + e);
					e.printStackTrace();
				}
				quit = (String)variables.get("quit");
			}
		}
		
		// "out" MUST be closed here, otherwise the file deletions in the below
		//  code block will not work
		out.close();
		
	}
}
