/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands;

import java.util.HashMap;
import java.util.Stack;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.Image;
import javax.tools.diagnostics.image.ImageAddressSpace;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.image.MemoryAccessException;

import org.apache.kato.katoview.Output;
import org.apache.kato.katoview.commands.helpers.Utils;


public class HexdumpCommand extends Command {

	public HexdumpCommand(Output o){
		super(o, "hexdump", "outputs a section of memory in a hexdump-like format", 
				"parameters: <hex_address> <bytes_to_print>\n\n" +
				"outputs <bytes_to_print> bytes of memory contents " +
				"starting from <hex_address>"				
				);
		child_commands = null;
	}
	
	public void doCommand(Stack args, Image loadedImage, HashMap properties){
		StringBuffer stringBuffer = new StringBuffer();
		ImageAddressSpace imageAddressSpace = Utils._extractAddressSpace(loadedImage);
		long addressInDecimal = 0;
		int numBytesToPrint = 16*16; //dump 256 bytes by default
		int asciiIndex = 0;
		
		if (!args.isEmpty()){
			Long address = Utils.longFromString((String)args.pop());
			if(null == address){
				out.error("Specified address is invalid");
				return;
			}
			addressInDecimal = address.longValue();
		}
		else
		{
			out.error("\"hexdump\" requires at least an address parameter");
			return;
		}
		
		if (!args.isEmpty()){
			try{
				numBytesToPrint = Integer.parseInt((String)args.pop());
			}catch(NumberFormatException nfe){
				out.error("Specified length is invalid");
				return;
			}
		}
		
		ImagePointer imagePointerBase = imageAddressSpace.getPointer(addressInDecimal);
		
		String asciiChars = "";
		
		long i; 
		for (i = 0; i < numBytesToPrint; i++){
			ImagePointer imagePointer = imagePointerBase.add(i);
			//stringBuffer.append(Long.toHexString(addressInDecimal+i)+":\t");
			//stringBuffer.append(Long.toHexString(imagePointer.getAddress())+":\t");
			try {
				Byte byteValue = new Byte(imagePointer.getByteAt(0));
				asciiIndex = (int) byteValue.byteValue();
				if (asciiIndex < 0) {
					asciiIndex += 256;
				}
				
				String rawHexString = Integer.toHexString(byteValue.intValue());
				String fixedHexString = fixHexStringLength(rawHexString);
				
				String hexText = fixedHexString;
				if (0 == i%4){
					hexText = " "+hexText;
				}
				
				if (0 == i%16){
					hexText  = "\n"+Long.toHexString(imagePointer.getAddress())+":"+hexText;
					asciiChars = "  |";
				}
				
				stringBuffer.append(hexText);
				
				asciiChars += Utils.byteToAscii.substring(asciiIndex,asciiIndex+1);
				if (15 == i%16 && i != 0){
					asciiChars += "|";
					//System.out.println("****"+asciiChars);
					stringBuffer.append(asciiChars);  
				}
			}catch (MemoryAccessException e) {
				out.error("Address not in memory - 0x" 
						+ Long.toHexString(imagePointer.getAddress()));
				return;
			}catch (CorruptDataException e) {
				out.error("Dump data is corrupted");
				return;
			}
		}
		
		long undisplayedBytes = 16 - i%16; 
		if (16 != undisplayedBytes){
			stringBuffer.append(padSpace(undisplayedBytes, asciiChars));
		}
		stringBuffer.append("\n");
		out.println(new String(stringBuffer));
		
		/*properties.put(Utils.CURRENT_MEM_ADDRESS, 
				Long.toHexString(addressInDecimal+numBytesToPrint));*/
		properties.put(Utils.CURRENT_MEM_ADDRESS, new Long(addressInDecimal));
		properties.put(Utils.CURRENT_NUM_BYTES_TO_PRINT, new Integer(numBytesToPrint));
	}
	
	private String padSpace(long undisplayedBytes, String asciiChars){
		for (int i = 0; i < 2*undisplayedBytes + undisplayedBytes/4; i++){
			asciiChars = " " + asciiChars;
		}
		return asciiChars;
	}
	
	private String fixHexStringLength(String rawHexString){
		int length = rawHexString.length();
		if (1 == length){
			return ("0" + rawHexString);
		}
		else if (2 == length){
			return rawHexString;
		}
		else if (8 == length){
			return rawHexString.substring(6, 8);
		}
		else{
			return "ERROR fixHexStringLength";
		}
	}
	
}
