/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands.tomcat;

import java.util.HashMap;
import java.util.List;
import java.util.Stack;

import javax.tools.diagnostics.image.CorruptData;
import javax.tools.diagnostics.image.Image;
import javax.tools.diagnostics.image.DiagnosticException;
import javax.tools.diagnostics.runtime.java.JavaClass;
import javax.tools.diagnostics.runtime.java.JavaLocation;
import javax.tools.diagnostics.runtime.java.JavaObject;
import javax.tools.diagnostics.runtime.java.JavaRuntime;
import javax.tools.diagnostics.runtime.java.JavaStackFrame;
import javax.tools.diagnostics.runtime.java.JavaThread;

import org.apache.kato.katoview.Output;
import org.apache.kato.katoview.commands.Command;
import org.apache.kato.katoview.commands.helpers.JavaRuntimeVisitor;
import org.apache.kato.katoview.commands.helpers.KatoSearch;

/**
 * Retrieves the version of Tomcat in the dump, or displays an error.
 *
 */
public class TomcatJSPCommand extends Command {
	private static final String SHORT_DESCRIPTION = "Shows JSPS";
	private static final String COMMAND_NAME = "jsp";
	private static final String LONG_DESCRIPTION = "Reports tomcat's thread pool state.";
	
	// Look for this class.
	private static final String JSP_CLASS="org/apache/jasper/runtime/HttpJspBase"; 

	
	public TomcatJSPCommand(Output o) {
		super(o, COMMAND_NAME, SHORT_DESCRIPTION, LONG_DESCRIPTION);
		
		child_commands = null;
	}

	public void doCommand(Stack args, Image loadedImage, HashMap properties) {
		try {
			KatoSearch.visitRuntimes(loadedImage, new PrintJSP());
		} catch (DiagnosticException e) {
			e.printStackTrace();
		}
	}

	/**
	 * A visitor for finding the correct class and then return
	 *
	 */
	private class PrintJSP implements JavaRuntimeVisitor {
		boolean found = false;
		JavaRuntime runtime;
		
		public void visit(JavaRuntime runtime) throws DiagnosticException {
			found = false;
			this.runtime = runtime;
			List<JavaClass> jsps = KatoSearch.findClasses(runtime, JSP_CLASS);
			
			for (JavaClass jsp : jsps) {
				// Avoid the JSP superclass itself.
				if (JSP_CLASS.equals(jsp.getName())) {
					continue;
				}
					
				found = true;
				out.println("Found JSP:  "+ jsp.getName() + ", instances: ");
				
				List<JavaObject> instances = KatoSearch.findInstances(runtime, jsp);
				
				for (JavaObject instance : instances) {
					out.println("\t0x"+Long.toHexString(instance.getID().getAddress()));
				}
				out.println("\nJavaThread: ");
				
				List<JavaThread> threads = KatoSearch.findClassInThreads(runtime, jsp);
				
				for (JavaThread thread : threads) {					 
					out.println("\tJava stack trace for thread `" + thread.getName() + "'");
					out.println("\t=========================================================\n");
					
					for (Object nextFrame : thread.getStackFrames()) {
						if (nextFrame instanceof CorruptData) {
							continue;
						}
						
						JavaLocation location = ((JavaStackFrame) nextFrame).getLocation();
						
						if (location != null) {
							out.println("\t"+location);
						}
					}
					out.println("");
				}
				out.println("");
			}
			
			if (found == false) {
				out.println("Couldn't find any JSPs");				
			}
		}

		
	}
}