/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands.tomcat;

import java.util.HashMap;
import java.util.List;
import java.util.Stack;

import javax.tools.diagnostics.image.Image;
import javax.tools.diagnostics.image.DiagnosticException;
import javax.tools.diagnostics.runtime.java.JavaClass;
import javax.tools.diagnostics.runtime.java.JavaField;
import javax.tools.diagnostics.runtime.java.JavaObject;
import javax.tools.diagnostics.runtime.java.JavaRuntime;

import org.apache.kato.katoview.Output;
import org.apache.kato.katoview.commands.Command;
import org.apache.kato.katoview.commands.helpers.JavaClassVisitor;
import org.apache.kato.katoview.commands.helpers.JavaRuntimeVisitor;
import org.apache.kato.katoview.commands.helpers.KatoSearch;

/**
 * Retrieves the version of Tomcat in the dump, or displays an error.
 *
 */
public class TomcatThreadPoolsCommand extends Command {
	private static final String SHORT_DESCRIPTION = "Shows thread pool state";
	private static final String COMMAND_NAME = "pools";
	private static final String LONG_DESCRIPTION = "Reports tomcat's thread pool state.";
	
	// Look for this class.
	private static final String THREAD_POOL_CLASS="org/apache/tomcat/util/threads/ThreadPool"; 
	private static final String POOL_NAME = "name",
								CURRENT_BUSY = "currentThreadsBusy",
								CURRENT_COUNT = "currentThreadCount";
	
	public TomcatThreadPoolsCommand(Output o) {
		super(o, COMMAND_NAME, SHORT_DESCRIPTION, LONG_DESCRIPTION);
		
		child_commands = null;
	}

	public void doCommand(Stack args, Image loadedImage, HashMap properties) {
		try {
			KatoSearch.visitRuntimes(loadedImage, new PrintVersion());
		} catch (DiagnosticException e) {
			e.printStackTrace();
		}
	}

	/**
	 * A visitor for finding the correct class and then return
	 *
	 */
	private class PrintVersion implements JavaRuntimeVisitor, JavaClassVisitor {
		boolean found = false;
		JavaRuntime runtime;
		
		public void visit(JavaRuntime runtime) throws DiagnosticException {
			found = false;
			this.runtime = runtime;
			KatoSearch.visitClasses(runtime, this);
			
			if (found == false) {
				out.println("ERROR: Couldn't find a Tomcat server");				
			}
		}

		public boolean visit(JavaClass clazz) throws DiagnosticException {
			if (clazz.getName().equals(THREAD_POOL_CLASS)) {
				found = true;
				JavaField nameField = KatoSearch.getField(clazz, POOL_NAME);
				JavaField busyField = KatoSearch.getField(clazz, CURRENT_BUSY);
				JavaField countField = KatoSearch.getField(clazz, CURRENT_COUNT);
				
				
				List<JavaObject> instances = KatoSearch.findInstances(runtime, clazz);
				if (instances.size() >0) {
					out.println("Found thread pools");
					for (JavaObject obj : instances) {
						try {
							out.println("Pool name:" + nameField.getString(obj)+", threads: "
									+ countField.getInt(obj) +
									", busy: " + busyField.getInt(obj));
						
						}catch (DiagnosticException e) {
							e.printStackTrace();
						}
					}
				}
				
				return false; // Stop searching through classes.
			}
			return true;
		}
		
	}
}