/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands.xcommands;

import java.util.HashMap;
import java.util.Stack;
import java.util.Iterator;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.Image;
import javax.tools.diagnostics.runtime.ManagedRuntime;
import javax.tools.diagnostics.runtime.java.JavaClass;
import javax.tools.diagnostics.runtime.java.JavaHeap;
import javax.tools.diagnostics.runtime.java.JavaObject;
import javax.tools.diagnostics.runtime.java.JavaRuntime;

import org.apache.kato.katoview.Output;
import org.apache.kato.katoview.commands.Command;
import org.apache.kato.katoview.commands.helpers.ClassOutput;
import org.apache.kato.katoview.commands.helpers.Exceptions;
import org.apache.kato.katoview.commands.helpers.Utils;


public class XJCommand extends Command {

	public XJCommand(Output o){
		super(o, "j", "displays information about a particular object or all " +
				"objects of a class",
				
				"parameters: 0x<object_addr>, <class_name>\n\n" +
				"If given class name, all static fields with their values will be " +
				"printed, followed by all objects of that class with their fields " +
				"and values.\n\n" +
				"If given an object address (in hex), static fields for that object's " +
				"class will not be printed; the other fields and values of that object " +
				"will be printed along with its address.\n\n" +
				"Note: this command ignores the number of items and unit size passed " +
				"to it by the \"x/\" command.\n"
		);
		child_commands = null;
	}
	
	public void doCommand(Stack args, Image loadedImage, HashMap properties)
	{
		args.pop();	// pop and discard number of units to display
		args.pop();	// pop and discard unit size
		
		boolean supers = false;
		
		if (args.isEmpty())
		{
			out.error("\"x/j\" command requires one or two parameters");
			return;
		}
		else
		{
			String param = (String)args.pop();
			Long objAddress;
			String objName;
			
			objAddress = Utils.longFromStringWithPrefix(param);
			if (null == objAddress)
			{
				objName = param;
				
				if (!args.isEmpty())
				{
					String option = (String)args.pop();
					if (option.equals("super"))
					{
						supers = true;
					}
					else if (option.equals("nosuper"))
					{
						supers = false;
					}
					else
					{
						out.error("second parameter for \"x/j\" must be \"super\" or" +
								" \"nosuper\"");
						return;
					}
				}
			}
			else
			{
				objName = null;
			}
			
			if (!args.isEmpty())
			{
				out.error("\"x/j\" takes at most one parameter if address is specified" +
						" or two parameters if class name is specified");
				return;
			}
			
			printRuntimeObjects(loadedImage, objAddress, objName, out, supers);
		}
	}
	
	private void printRuntimeObjects(Image loadedImage, Long objAddress, String objName,
			Output out, boolean supers)
	{
		ManagedRuntime mr;
		Iterator itRuntime = Utils.getRuntimes(loadedImage);
		int count = 1;

		out.print("\n");
		while (itRuntime.hasNext()) {
			mr = (ManagedRuntime)itRuntime.next();
			if (mr instanceof JavaRuntime)
			{
				out.print("\truntime #" + count + " - version: ");
				try {
					out.print(mr.getVersion());
				} catch (CorruptDataException e) {
					out.print(Exceptions.getCorruptDataExceptionString());
				}
				out.print("\n");
				
				printHeapObjects((JavaRuntime)mr, objAddress, objName, out, supers);
			}
			count++;
		}
	}
	
	private void printHeapObjects(JavaRuntime jr, Long objAddress, String objName,
			Output out, boolean supers)
	{
		JavaHeap jh;
		Iterator itHeap = jr.getHeaps().iterator();
		int count = 1;
		
		while (itHeap.hasNext()) {
			jh = (JavaHeap)itHeap.next();
			
			out.print("\t heap #" + count + " - name: ");
			out.print(jh.getName());
			out.print("\n\n");
			
			printObjects(jh, objAddress, objName, out, supers, jr);
			count++;
		}
	}
	
	private void printObjects(JavaHeap jh, Long objAddress, String objName, Output out,
			boolean supers, JavaRuntime jr)
	{
		JavaObject jo;
		Iterator itObject = jh.getObjects().iterator();
		boolean found = false;

		if (objName != null)
		{
			JavaClass objClass = Utils.getClassGivenName(objName, jr, out);
			
			// if we couldn't find a class of that name, return; the passed in class name could
			//  still be an array type or it might not exist
			if (null == objClass)
			{
				out.print("\t  could not find class with name \"" + objName + "\"\n\n");
				return;
			}
			
			ClassOutput.printStaticFields(objClass, out);
			
			while (itObject.hasNext())
			{
				jo = (JavaObject)itObject.next();


				String className;
				String hierarchy = "";
				JavaClass jc;

				try {
					jc = jo.getJavaClass();
				} catch (CorruptDataException e) {
					out.print("\t  <error getting class while traversing objects: ");
					out.print(Exceptions.getCorruptDataExceptionString());
					out.print(">\n");
					jc = null;
				}

				boolean foundSuperclass = false;
				while (jc != null && !foundSuperclass)
				{
					try {
						className = jc.getName();
					} catch (CorruptDataException e) {
						out.print("\t  <error getting class name while traversing objects: ");
						out.print(Exceptions.getCorruptDataExceptionString());
						out.print(">\n");
						jc = null;
						continue;
					}
						
					if (hierarchy.equals(""))
						hierarchy = className;
					else
						hierarchy = className + " => " + hierarchy;

					if (className.equals(objName))
					{
						found = true;
						foundSuperclass = true;
						out.print("\t  ");
						out.print(hierarchy);
						out.print(" @ ");
						out.print(Utils.toHex(jo.getID().getAddress()));
						out.print("\n");
						
						ClassOutput.printFields(jo, jc, jr, out);
					}
					else
					{
						if (supers)
						{
							try {
								jc = jc.getSuperclass();
							} catch (CorruptDataException e) {
								out.print("\t  <error getting superclass while traversing objects: ");
								out.print(Exceptions.getCorruptDataExceptionString());
								out.print(">\n");
								jc = null;
							}
						}
						else
						{
							jc = null;
						}
					}
				}
			}
		}
		else // objAddress != null
		{
			boolean done = false;
			while (itObject.hasNext() && !done)
			{
				jo = (JavaObject)itObject.next();


				if (jo.getID().getAddress() == objAddress.longValue())
				{
					JavaClass jc;
					
					found = true;
					out.print("\t  ");
					try {
						jc = jo.getJavaClass();
					} catch (CorruptDataException e) {
						out.print("\t  <error getting class while traversing objects: ");
						out.print(Exceptions.getCorruptDataExceptionString());
						out.print(">");
						jc = null;
					}
					
					if (null != jc)
					{
						try {
							out.print(jc.getName());
						} catch (CorruptDataException e) {
							out.print("\t  <error getting class name while traversing objects: ");
							out.print(Exceptions.getCorruptDataExceptionString());
							out.print(">");
						}
						out.print(" @ ");
						out.print(Utils.toHex(objAddress.longValue()));
						out.print("\n");
						
						ClassOutput.printFields(jo, jc, jr, out);
					
						done = true;	// assumes only one object can exist at a specific memory address
					}
				}
			}
		}
		
		if (!found)
		{
			if (objName != null)
			{
				out.print("\t  <no object of class \"");
				out.print(objName);
				out.print("\" exists>\n\n");
			}
			else // objAddress != null
			{
				out.print("\t  <no object found at address ");
				out.print(Utils.toHex(objAddress.longValue()));
				out.print(">\n\n");
			}
		}
	}

}
