/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands.xcommands;

import java.util.HashMap;
import java.util.Stack;
import java.util.Iterator;

import javax.tools.diagnostics.image.CorruptData;
import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.Image;
import javax.tools.diagnostics.image.ImageAddressSpace;
import javax.tools.diagnostics.image.ImageModule;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.image.ImageProcess;
import javax.tools.diagnostics.image.ImageSection;
import javax.tools.diagnostics.image.ImageSymbol;
import javax.tools.diagnostics.image.MemoryAccessException;

import org.apache.kato.katoview.Output;
import org.apache.kato.katoview.commands.Command;
import org.apache.kato.katoview.commands.helpers.Exceptions;
import org.apache.kato.katoview.commands.helpers.Utils;


public class XKCommand extends Command {

	public XKCommand(Output o){
		super(o, "k", "displays the specified memory section as if it were a stack frame",
				"parameters: 0x<addr>\n\n" +
				"Displays the value of each section (whose size is defined by the " +
				"pointer size of this architecture) of memory, adjusted for the " +
				"endianness of the architecture this dump file is from, starting at the " +
				"specified address.  It also displays a module with a module section " +
				"and an offset from the start of that module section in memory if the " +
				"pointer points to that module section.  " +
				"If no symbol is found, it displays a \"*\" and an offset from the " +
				"current address if the pointer points to an address within 4KB (4096 " +
				"bytes) of the current address.\n\n" +
				"While this command can work on an arbitrary section of memory, it is " +
				"probably most useful when used on a section of memory that refers to a " +
				"stack frame.  To find the memory section of a thread's stack frame, use " +
				"the \"info thread\" command.\n\n" +
				"Note: This command uses the number of items passed to " +
				"it by the \"x/\" command, but ignores the unit size.\n");
		child_commands = null;
	}
	
	public void doCommand(Stack args, Image loadedImage, HashMap properties)
	{
		int numUnits = ((Integer)args.pop()).intValue();
		args.pop();	// pop and discard unitSize; we ignore this value
		
		if (args.isEmpty())
		{
			out.error("\"x/k\" command takes exactly one parameter");
			return;
		}

		String param = (String)args.pop();
		if (!args.isEmpty())
		{
			out.error("\"x/k\" only takes one parameter");
			return;
		}
		
		Long address = Utils.longFromStringWithPrefix(param);
		
		if (null == address)
		{
			out.error("invalid hex address specified; address must be specified as "
					+ "\"0x<hex_address>\"");
			return;
		}
		
		ImageAddressSpace ias = (ImageAddressSpace)properties.get("current_address_space");
		int pointerSize = getIASPointerSize(ias);
		int unitSize;
		
		if (pointerSize > 32)
			unitSize = 8;
		else
			unitSize = 4;
		
		out.print("\n");
		
		for (int index = 0; index < numUnits; index++)
		{
			boolean found = false;
			long currAddr = address.longValue() + (index * unitSize);
			ImagePointer ip = ias.getPointer(currAddr);
			
			out.print("\t");
			out.print(Utils.toHex(currAddr));
			out.print("   ");

			long l = 0;
			try {
				l = ip.getPointerAt(0).getAddress();
				found = true;
			} catch (CorruptDataException e) {
				found = false;
			} catch (MemoryAccessException e) {
				found = false;
			}

			if (found) {
				long pointer = l;
				
				out.print(toAdjustedHex(l, pointerSize));
				out.print("   ");

				if (31 == pointerSize) {
					pointer = (int)(pointer & (((long)1) << pointerSize) - 1);
				}
				
				if (printSymbol(pointer, l - currAddr, pointerSize, ias)) {
				} else if (printStackPointer(pointer, l - currAddr, pointerSize, ias)) {
				}
				out.print("\n");
			} else {
				out.print("<address not found in any address space or exception occurred>\n");
			}
		}
		
		out.print("\n");
	}
	
	private boolean printSymbol(long pointer, long diff, int pointerSize, ImageAddressSpace ias)
	{
		Iterator itProcess = ias.getProcesses().iterator();
		while (itProcess.hasNext()) {
			ImageProcess ip = (ImageProcess)itProcess.next();
			Iterator itModule;
			try {
				itModule = ip.getLibraries().iterator();
			} catch (CorruptDataException e) {
				// FIXME
				itModule = null;
			} catch (DataUnavailable e) {
				// FIXME
				itModule = null;
			}
			while (null != itModule && itModule.hasNext()) {
				ImageModule im = (ImageModule)itModule.next();
				Iterator itImageSection = im.getSections().iterator();
				while (itImageSection.hasNext()) {
					ImageSection is = (ImageSection)itImageSection.next();
					long startAddr = is.getBaseAddress().getAddress();
					long endAddr = startAddr + is.getSize();
					
					if (pointer >= startAddr && pointer < endAddr) {
						/* can we find a matching symbol? */
						long maxDifference = pointer - startAddr;
						ImageSymbol bestSymbol = null;
						for (Iterator iter = im.getSymbols().iterator(); iter.hasNext();) {
							Object next = iter.next();
							if (next instanceof CorruptData)
								continue;
							ImageSymbol symbol = (ImageSymbol) next;
							long symbolAddress = symbol.getAddress().getAddress();
							if (symbolAddress <= pointer && pointer - symbolAddress < maxDifference) {
								maxDifference = pointer - symbolAddress;
								bestSymbol = symbol;
							}
						}

						try {
							out.print(im.getName());
						} catch (CorruptDataException e) {
							out.print(Exceptions.getCorruptDataExceptionString());
						}
						out.print("::");
						if (bestSymbol == null) {
							out.print(is.getName());
						} else {
							out.print(bestSymbol.getName());
						}
						out.print("+");
						out.print(Long.toString(maxDifference));
						
						// if we find the address in the symbols, there's no need to continue
						//  trying to find it elsewhere in the address space
						return true;
					}
				}
			}
		}
		
		return false;
	}
	
	private boolean printStackPointer(long pointer, long diff, int pointerSize, ImageAddressSpace ias)
	{
		if (Math.abs(diff) <= 4096) {	
			if (diff < 0) {
				out.print(" .");
				out.print(Long.toString(diff));
			} else {
				out.print(" .+");
				out.print(Long.toString(diff));
			}
			return true;
		}
		return false;
	}
	
	private int getIASPointerSize(ImageAddressSpace ias)
	{
		return ((ImageProcess)ias.getProcesses().iterator().next()).getPointerSize();
	}

	private String toAdjustedHex(long l, int pointerSize)
	{
		if (pointerSize > 32) {
			return Utils.toFixedWidthHex(l);
		} else if (31 == pointerSize) {
			return Utils.toFixedWidthHex((int)(l & (((long)1) << pointerSize) - 1));
		} else {
			return Utils.toFixedWidthHex((int)l);
		}
	}
}
